#include "../../internal-include/CD/cd_internal.hpp"
#include <PSX/System/IOPorts/dma_io.hpp>
#include <PSX/System/IOPorts/interrupt_io.hpp>
#include <PSX/System/syscalls.hpp>

namespace JabyEngine {
    namespace CD {
        namespace internal {
            static constexpr auto AudioSectorMode = CD_IO::Mode::from(CD_IO::Mode::SingleSpeed, CD_IO::Mode::CDDA);
            static constexpr auto DataSectorMode  = CD_IO::Mode::from(CD_IO::Mode::DoubleSpeed, CD_IO::Mode::DataSector);
            
            static SysCall::InterruptVerifierResult interrupt_verifier();
            static uint32_t interrupt_handler(uint32_t);

            static SectorBufferAllocator sector_allocator;
            static uint32_t              cur_lba;
            static uint32_t              dst_lba;
            
            CD_IO::Interrupt::Type    last_interrupt    = CD_IO::Interrupt::Type::None;
            uint8_t                   cmd_interrupt_bit = 0;
            State                     current_state     = State::Free;
            SysCall::InterrupCallback callback          = {
                .next              = nullptr,
                .handler_function  = interrupt_handler,
                .verifier_function = interrupt_verifier
            };

            static void pause_cd() {
                CD_IO::PortIndex0::change_to();
                Command::send<CD_IO::PortIndex0>(CD_IO::Command::Pause);
            }

            // Requires Index0
            static void read_cd(uint32_t lba) {
                const auto loc = CDTimeStamp::from(lba);               

                Command::send_wait<CD_IO::PortIndex0>(CD_IO::Command::SetLoc, loc.get_min_cd(), loc.get_sec_cd(), loc.get_sector_cd());
                Command::send<CD_IO::PortIndex0>(CD_IO::Command::ReadN);
                current_state = State::Reading;
            }

            static void read_sector_dma(CD_IO::DataSector& sector) {
                static const auto WaitSectorReady = []() {
                    while(!CD_IO::IndexStatus.read().is_set(CD_IO::IndexStatus::HasDataFifoData));
                };

                static const auto ReadSector = [](uint32_t* dst) {
                    DMA_IO::CDROM.set_adr(reinterpret_cast<uintptr_t>(dst));
                    DMA_IO::CDROM.block_ctrl.write(DMA_IO::BCR::SyncMode0::for_cd());
                    DMA_IO::CDROM.channel_ctrl.write(DMA_IO::CHCHR::StartCDROM());
                    
                    DMA_IO::CDROM.wait();

                    CD_IO::PortIndex0::Request.write(CD_IO::Request::reset());
                };

                WaitSectorReady();
                ReadSector(sector.data);
            }

            static void read_sector_to(CD_IO::DataSector& sector) {
                CD_IO::PortIndex0::change_to();
                CD_IO::PortIndex0::Request.write(CD_IO::Request::want_data());

                // We only support DMA rn
                read_sector_dma(sector);

                // Do we ever want to support reading via IO Port?
                // Doesn't seem to important when we can use DMA
            }

            static SysCall::InterruptVerifierResult interrupt_verifier() {
                if(Interrupt::is_irq(Interrupt::CDROM)) {
                    return SysCall::InterruptVerifierResult::ExecuteHandler;
                }
                
                else {
                    return SysCall::InterruptVerifierResult::SkipHandler;
                }
            }

            static uint32_t interrupt_handler(uint32_t) {
                const auto old_status = CD_IO::IndexStatus.read();

                CD_IO::PortIndex1::change_to();
                const auto cur_irq = CD_IO::Interrupt::get_type(CD_IO::PortIndex1::InterruptFlag);
                last_interrupt = cur_irq;
                CD_IO::Interrupt::ack_extended(CD_IO::PortIndex1::InterruptFlag);

                cmd_interrupt_bit = bit::clear(cmd_interrupt_bit, cur_irq);
                if(cur_irq == CD_IO::Interrupt::DataReady) {
                    // Obtain sector content here
                    auto* sector = sector_allocator.allocate_sector();
                    if(sector) {
                        //Now obtain sector
                        read_sector_to(*sector);

                        cur_lba++;
                        if(cur_lba == dst_lba) {
                            current_state = State::Done;
                            pause_cd();
                        }
                    }

                    else {
                        current_state = State::BufferFull;
                        pause_cd();
                    }
                }

                else if(cur_irq == CD_IO::Interrupt::DiskError) {
                    current_state = State::Error;
                }

                // No masking required because we can only write bit 0 - 2
                CD_IO::IndexStatus.write(old_status);
                Interrupt::ack_irq(Interrupt::CDROM);
                SysCall::ReturnFromException();
                __builtin_unreachable();
            }

            void read_file(AutoLBAEntry file_info, const SectorBufferAllocator& buffer_allocator) {
                cur_lba          = file_info.get_lba();
                dst_lba          = cur_lba + file_info.get_size_in_sectors();
                sector_allocator = buffer_allocator;

                Command::wait_completed();
                CD_IO::PortIndex0::change_to();
                Command::send_wait<CD_IO::PortIndex0>(CD_IO::Command::SetMode, DataSectorMode);

                read_cd(cur_lba);
            }

            void continue_reading() {
                if(current_state == State::BufferFull) {
                    Command::wait_completed();
                    read_cd(cur_lba);    
                }                
            }

            void enable_CDDA() {
                Command::wait_completed();
                CD_IO::PortIndex0::change_to();
                Command::send_wait<CD_IO::PortIndex0>(CD_IO::Command::SetMode, AudioSectorMode);                
            }
        }
    }
}